//--------------------------------------------------------------------------------------
// File: fogMesh.fx
//
// The effect file for OceanFog2.  
// 
//--------------------------------------------------------------------------------------

#define magnitude(v) sqrt(v.x*v.x+v.y*v.y+v.z*v.z)
#define falloff(m,d) (m/abs(d)-1) //hyperbolic falloff

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
cbuffer cb0
{
	float4 g_MaterialAmbientColor;      // Material's ambient color
	float3 g_LightDir;                  // Light's direction in world space
	float4 g_LightDiffuse;              // Light's diffuse color
	float3 g_CameraPos;					// Camera position
	float3 g_CameraForward;				// Camera forward vector
	float  g_fNormalMapFactor;			// Intensity of normal map

	float    g_fTime;                   // App's time in seconds
	float4x4 g_mWorld;                  // World matrix for object
	float4x4 g_mWorldViewProjection;    // World * View * Projection matrix	

	float3	 g_vSpotDir;				// spotlight direction
	float3	 g_vSpotPos;				// spotlight position
	float	 g_fSpotFrustum;			// spotlight frustum angle
	float	 g_fSpotIntensity;			// spotlight intensity multiplier
	float4	 g_vSpotDiffuse;			// spotlight diffuse color
	float4x4 g_mSpotWorld;				// world matrix for spotlight
	
	bool	underwater;	
	float   g_FogDensity;
	
	float4	ClipPlane;
}

Texture3D<float> g_FogTexture;			// Fog texture for mesh
Texture2D g_NormalMap;
Texture2D g_MeshTexture;				// Color texture for mesh

//--------------------------------------------------------------------------------------
// Texture samplers
//--------------------------------------------------------------------------------------
SamplerState samFog
{
    Filter = MIN_MAG_MIP_LINEAR ;
    AddressU = Wrap;
    AddressV = Wrap;
    AddressW = Wrap;
};

SamplerState samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};

//--------------------------------------------------------------------------------------
// Blend states
//--------------------------------------------------------------------------------------
BlendState SrcAlphaBlendingOff
{
    BlendEnable[0] = FALSE;
};

//--------------------------------------------------------------------------------------
// Vertex shader output structure
//--------------------------------------------------------------------------------------
struct RenderSceneVSOUT
{
    float4 Position				: SV_POSITION0;   // transformed vertex position 
	float  ClipDistance			: SV_CLIPDISTANCE;
    float3 FogPos				: POSITION1;  // transformed fog position
    float3 WorldPos				: POSITION2;  // transformed world position
    float4 CameraDist			: POSITION3;		// distance from vertex to camera (magnitude in CameraDist.a)
    float4 Diffuse				: COLOR0;     // vertex diffuse color (note that COLOR0 is clamped from 0..1)
    float2 TextureUV			: TEXCOORD0;  // vertex texture coords 
    float3 vNormalWorldSpace	: POSITION4;
};

//--------------------------------------------------------------------------------------
// Shader Functions
//--------------------------------------------------------------------------------------

// Gets the fog coefficient f
float4 CalcFogCoeff( float4 SVPos, float3 FogPos, float3 WorldPos ) : SV_TARGET
{
	float CameraDist;
    CameraDist = SVPos.z; //distance( g_CameraPos, WorldPos );

    // Calculate fog based on perlin noise texture (g_FogTexture)
	float4 fog = g_FogTexture.Sample(samFog, FogPos);
	
	// height effect
	fog *= (g_CameraPos.y-WorldPos.y-FogPos.y)*(g_CameraPos.y-WorldPos.y-FogPos.y)*0.5f + 20.0f*CameraDist;
    
    return exp(-(g_FogDensity*fog*CameraDist)); // f=e^(-(density*fog*distance)^2
}

float4 CalcWaterCoeff( float3 WorldPos )
{
	float CameraDist;
    CameraDist = distance( g_CameraPos, WorldPos );

	return  exp(-(0.1f*CameraDist));
}

RenderSceneVSOUT VS( float4 vPos : POSITION, 
                         float3 vNormal : NORMAL,
                         float2 vTexCoord0 : TEXCOORD )
{
    RenderSceneVSOUT Output;
    //vPos.y -= 1.0f;
    
    // Transform the position from object space to homogeneous projection space
    Output.Position = mul(vPos, g_mWorldViewProjection);
    Output.WorldPos = mul(vPos, g_mWorld);
    
    // Transform the normal from object space to world space    
    Output.vNormalWorldSpace = normalize(mul(vNormal, (float3x3)g_mWorld)); // normal (world space)
    
    // Just copy the texture coordinate through
    Output.TextureUV = vTexCoord0; 

	Output.ClipDistance.x = dot( float4( Output.WorldPos, 1.0f ), ClipPlane );

    return Output;
}

float4 PS( RenderSceneVSOUT In ) : SV_TARGET
{ 
	float4 vWaterColor = { 0.02f, 0.025f, 0.035f, 1.0f };
    
    // normal map
    In.vNormalWorldSpace = normalize( In.vNormalWorldSpace + g_fNormalMapFactor*g_NormalMap.Sample( samLinear, In.TextureUV ) );
    
    // Calc diffuse color
	In.Diffuse.rgb = g_LightDiffuse * saturate(dot(In.vNormalWorldSpace, g_LightDir));		
	In.Diffuse.rgb += g_MaterialAmbientColor;  
    In.Diffuse.a = 1.0f;
    
    // Actual mesh's texture
    float3 transform = {-1.0f, 1.0f, -1.0f};
    In.TextureUV *= transform;
    In.Diffuse = g_MeshTexture.Sample( samLinear, In.TextureUV ) * In.Diffuse; // actual texture
	
    // spotlight
	float3 distanceVec = In.WorldPos - g_vSpotPos;
	float3 SpotWorldDir = mul(g_vSpotDir, g_mSpotWorld);
	float theta = acos( dot( normalize(distanceVec), normalize(SpotWorldDir) ) );
	if( abs(theta) < g_fSpotFrustum )
	{
		float SpotFalloff = saturate(falloff(g_fSpotFrustum,theta));
		if( underwater )
			In.Diffuse.rgb += saturate(-dot( SpotWorldDir, In.vNormalWorldSpace))*vWaterColor*SpotFalloff*g_fSpotIntensity*0.08f;
		else
			In.Diffuse.rgb += saturate(-dot( SpotWorldDir, In.vNormalWorldSpace))*g_vSpotDiffuse*SpotFalloff*g_fSpotIntensity*0.08f;
	}
	
	if( underwater )
	{
		float f = CalcWaterCoeff(In.WorldPos);
		return (In.Diffuse*f + vWaterColor*(1-f));
	}
    
    // fog coefficient
    float3 vFogDirection = {0.0f, 0.0f, -1.0f}; // for animation
    float4 vFogColor = { 0.6f, 0.6f, 0.6f, 1.0f };
    float f = abs(CalcFogCoeff(In.Position, In.WorldPos/40.0f + (g_fTime*0.05f)*vFogDirection, In.WorldPos));
    
    // Mesh texture multiplied with fog
    return (In.Diffuse*f + vFogColor*(1-f)); // finalcolor = originalcolor*f + fogcolor*(1-f)
}

RenderSceneVSOUT VSlamp( float4 vPos : POSITION, 
                         float3 vNormal : NORMAL,
                         float2 vTexCoord0 : TEXCOORD )
{
    RenderSceneVSOUT Output;
    
    // Transform the position from object space to homogeneous projection space
    Output.Position = mul(vPos, g_mWorldViewProjection);

	Output.ClipDistance.x = dot( float4( Output.WorldPos, 1.0f ), ClipPlane );

    return Output;
}

float4 PSlamp( RenderSceneVSOUT In ) : SV_TARGET
{ 
    return float4(1,1,1,1);
}

//--------------------------------------------------------------------------------------
// Renders scene 
//--------------------------------------------------------------------------------------
technique10 RenderScene
{
    pass P0
    {       
		SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS() ) );
        
        SetBlendState( SrcAlphaBlendingOff, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
	}
}

technique10 RenderLamp
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VSlamp() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSlamp() ) );
        
        SetBlendState( SrcAlphaBlendingOff, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
	}
}

