//--------------------------------------------------------------------------------------
// File: spotvolume.fx
// 
// volumetric foggy light
//--------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
cbuffer cb0
{
	float3 g_LightDir;                  // Light's direction in world space
	float4 g_LightDiffuse;              // Light's diffuse color
	float3 g_CameraPos;					// Camera position
	float4 g_fSpotDiffuse;
	float  g_fSpotIntensity;
	float3 g_vSpotCenter;
	
	float  g_fFogDensity;
	float4 g_vfogColor = { 0.6f, 0.6f, 0.6f, 1.0f }; 

	float    g_fTime;                   // App's time in seconds
	float4x4 g_mWorld;                  // World matrix for object
	float4x4 g_mWorldViewProjection;    // = World * View * Projection matrices	
	bool	 underwater;
}

Texture3D<float> g_FogTexture;			// Fog texture for mesh

//--------------------------------------------------------------------------------------
// Texture samplers
//--------------------------------------------------------------------------------------
SamplerState samFog
{
    Filter = ANISOTROPIC ;
    AddressU = Wrap;
    AddressV = Wrap;
    AddressW = Wrap;
};

//--------------------------------------------------------------------------------------
// Blend states
//--------------------------------------------------------------------------------------
BlendState SrcAlphaBlendingAdd
{
    BlendEnable[0] = TRUE;
    SrcBlend = SRC_ALPHA;
    DestBlend = ONE;
    BlendOp = ADD;
    SrcBlendAlpha = ZERO;
    DestBlendAlpha = ZERO;
    BlendOpAlpha = ADD;
    RenderTargetWriteMask[0] = 0x0F;
};

//--------------------------------------------------------------------------------------
// Vertex shader output structure
//--------------------------------------------------------------------------------------
struct RenderSceneVSOUT
{
    float4 Position   : SV_POSITION0;   // transformed vertex position 
    float3 FogPos     : POSITION1;  // transformed fog position
    float3 WorldPos   : POSITION2;  // transformed world position
    float  SpotlightDist : DEPTH;		// spotlight distance for falloff
    float4 Diffuse    : COLOR0;     // vertex diffuse color (note that COLOR0 is clamped from 0..1)
    float2 TextureUV  : TEXCOORD0;  // vertex texture coords 
};

//--------------------------------------------------------------------------------------
// Shader Functions
//--------------------------------------------------------------------------------------

// Gets the fog coefficient f
float4 CalcFogCoeff( float4 SVPos, float3 fogPos, float density ) : SV_TARGET
{
	float CameraDist = SVPos.z;
	
    // Calculate fog based on perlin noise texture (g_FogTexture)
	float4 fog = g_FogTexture.Sample(samFog, fogPos);
    
    return exp(-(density*fog*CameraDist)); // f=e^(-(density*fog*distance)^2
}

float4 CalcWaterCoeff( float4 SVPos )
{
	float CameraDist;
    CameraDist = SVPos.z;

	return  exp(-(0.1f*CameraDist));
}

RenderSceneVSOUT VS( float4 vPos : POSITION )
{
	RenderSceneVSOUT Output;
	
    float3 FogDirection = {0.0f, 0.0f, -1.0f}; // for animation
    
    // Transform the position from object space to homogeneous projection space
    Output.Position = mul(vPos, g_mWorldViewProjection);
    Output.WorldPos = mul(vPos, g_mWorld);
    
    Output.SpotlightDist = distance(Output.WorldPos, g_vSpotCenter);

    Output.FogPos = Output.WorldPos; // vertex world position
    Output.FogPos /= 40.0f;
    Output.FogPos += (g_fTime*0.05f)*FogDirection; // animation
		
	return Output;
}

float4 PS( RenderSceneVSOUT Input ) : SV_TARGET
{
	if( underwater )
	{
		float4 vWaterColor = { .02f, .025f, .035f, 1.0f };
		float f = CalcWaterCoeff( Input.Position );
		return vWaterColor*(1-f) * g_fSpotIntensity * saturate(5/Input.SpotlightDist) * g_fSpotDiffuse * 25.0f;
	} 
	
	float f = CalcFogCoeff( Input.Position, Input.FogPos, g_fFogDensity );
	
	return g_fSpotIntensity * g_vfogColor*(1-f) * saturate(5/Input.SpotlightDist * g_fSpotDiffuse * 200.0f) * 2;
}

//--------------------------------------------------------------------------------------
// Renders Light Sphere 
//--------------------------------------------------------------------------------------
technique10 RenderSpotVolume
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS() ) );
        
        SetBlendState( SrcAlphaBlendingAdd, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
	}
}